import warnings
from pathlib import Path
from uuid import uuid1

import cryohub
import numpy as np
import pandas as pd
from cryotypes.image import ImageProtocol
from cryotypes.poseset import PoseSetProtocol
from scipy.spatial.transform import Rotation

from .utils import generate_vectors, invert_xyz


def get_reader(path):
    return read_layers


def _construct_positions_layer(
    coords, features, scale, exp_id, p_id, source, **pt_kwargs
):
    feat_defaults = (
        pd.DataFrame(features.iloc[-1].to_dict(), index=[0])
        if len(features)
        else pd.DataFrame()
    )
    feat_defaults["orientation"] = np.array(Rotation.identity(), dtype=object)
    return (
        coords,
        {
            "name": f"{exp_id} - particle positions",
            "features": features,
            "feature_defaults": feat_defaults,
            "face_color": "teal",
            "size": 5 / scale,
            "edge_width": 0,
            "scale": [scale] * 3,
            "shading": "spherical",
            "antialiasing": 0,
            "metadata": {"experiment_id": exp_id, "p_id": p_id, "source": source},
            # "projection_mode": "all",  # only available in napari 0.5.0
            "out_of_slice_display": True,
            **pt_kwargs,
        },
        "points",
    )


def _construct_orientations_layer(coords, features, scale, exp_id, p_id, source):
    if coords is None:
        vec_data = None
        vec_color = "blue"
    else:
        vec_data, vec_color = generate_vectors(
            invert_xyz(coords), features["orientation"]
        )
        vec_data = invert_xyz(vec_data)
    return (
        vec_data,
        {
            "name": f"{exp_id} - particle orientations",
            "edge_color": vec_color,
            "length": 100 / np.array(scale),
            "edge_width": 10 / np.array(scale),
            "scale": [scale] * 3,
            "metadata": {"experiment_id": exp_id, "p_id": p_id, "source": source},
            "vector_style": "arrow",
            # "projection_mode": "all",  # only available in napari 0.5.0
            "out_of_slice_display": True,
        },
        "vectors",
    )


def construct_particle_layer_tuples(
    coords,
    features,
    scale,
    exp_id,
    p_id=None,
    source="",
    **pt_kwargs,
):
    """
    Constructs particle layer tuples from particle data.

    Data is assumed to already by in zyx napari format, while features is
    the normal poseset dataframe.
    """
    # unique id so we can connect layers safely
    p_id = p_id if p_id is not None else uuid1()

    if features is None:
        features = pd.DataFrame()

    if "orientation" not in features.columns:
        features["orientation"] = np.array(
            [] if coords is None else Rotation.identity(len(coords)), dtype=object
        )

    # divide by scale top keep constant size. TODO: remove after vispy 0.12 which fixes this
    pos = _construct_positions_layer(
        coords, features, scale, exp_id, p_id, source, **pt_kwargs
    )
    ori = _construct_orientations_layer(coords, features, scale, exp_id, p_id, source)

    # ori should be last, or the auto-update feedback loop messes up the orientations
    # when existing layers are updated from a new run
    return [pos, ori]


def read_particles(particles):
    """Takes a valid poseset and converts it into napari layers."""
    # order is zyx in napari
    coords = invert_xyz(particles.position)

    if particles.features is not None:
        features = particles.features.copy(deep=False)
    else:
        features = pd.DataFrame()

    px_size = particles.pixel_spacing
    if not px_size:
        warnings.warn("unknown pixel spacing, setting to 1 Angstrom")
        px_size = 1

    if particles.shift is not None:
        shifts = invert_xyz(particles.shift)
        coords = coords + shifts
        shift_cols = ["shift_z", "shift_y", "shift_x"]
        features[shift_cols] = shifts
    if particles.orientation is not None:
        features["orientation"] = np.asarray(particles.orientation, dtype=object)

    return construct_particle_layer_tuples(
        coords, features, px_size, particles.experiment_id, particles.source
    )


def read_image(image):
    px_size = image.pixel_spacing
    if not px_size:
        warnings.warn("unknown pixel spacing, setting to 1 Angstrom")
        px_size = 1
    return (
        image.data,
        {
            "name": f"{image.experiment_id} - image",
            "scale": [px_size] * image.data.ndim,
            "metadata": {"experiment_id": image.experiment_id, "stack": image.stack},
            "interpolation2d": "spline36",
            "interpolation3d": "linear",
            "rendering": "average",
            "depiction": "plane",
            "blending": "translucent",
            "plane": {"thickness": 5},
            # "projection_mode": "mean",  # only available in napari 0.5.0
        },
        "image",
    )


def read_surface_picks(path):
    lines = []
    with open(path, "rb") as f:
        scale = np.load(f)
        surf_id = np.load(f)
        edge_color_cycle = np.load(f)
        while True:
            try:
                lines.append(np.load(f))
            except ValueError:
                break
        exp_id = f.read().decode()

    return (
        lines,
        {
            "name": f"{exp_id} - surface lines",
            "edge_width": 50 / scale[0],
            "metadata": {"experiment_id": exp_id},
            "scale": scale,
            "features": {"surface_id": surf_id},
            "feature_defaults": {"surface_id": surf_id.max() + 1},
            "edge_color_cycle": edge_color_cycle,
            "edge_color": "surface_id",
            "shape_type": "path",
            "ndim": 3,
        },
        "shapes",
    )


def read_surface(path):
    with open(path, "rb") as f:
        scale = np.load(f)
        # TODO: needs to exposed in napari
        # colormap = np.load(f)
        data = tuple(np.load(f) for _ in range(3))
        exp_id = f.read().decode()

    return (
        data,
        {
            "name": f"{exp_id} - surface",
            "metadata": {"experiment_id": exp_id},
            "shading": "smooth",
            "scale": scale,
            # TODO: needs to exposed in napari
            # colormap=colormap
        },
        "surface",
    )


def read_layers(*paths, **kwargs):
    layers = []
    cryohub_paths = []
    for path in paths:
        path = Path(path)
        if path.suffix == ".picks":
            layers.append(read_surface_picks(path))
        if path.suffix == ".surf":
            layers.append(read_surface(path))
        else:
            cryohub_paths.append(path)

    data_list = cryohub.read(*cryohub_paths, **kwargs)
    # sort so we get images first, better for some visualization circumstances
    for data in sorted(data_list, key=lambda x: not isinstance(x, ImageProtocol)):
        if isinstance(data, ImageProtocol):
            layers.append(read_image(data))
        elif isinstance(data, PoseSetProtocol):
            layers.extend(read_particles(data))

    for lay in layers:
        lay[1]["visible"] = False  # speed up loading
    return layers or None



import mrcfile
import numpy as np
import pandas as pd
import pytest
import starfile


@pytest.fixture(scope="session")
def star_file(tmp_path_factory):
    df = pd.DataFrame(
        {
            "rlnCoordinateX": [1, 1],
            "rlnCoordinateY": [2, 2],
            "rlnCoordinateZ": [3, 3],
            "rlnOriginX": [0.1, 0.1],
            "rlnOriginY": [0.2, 0.2],
            "rlnOriginZ": [0.3, 0.3],
            "rlnAngleRot": [0, 0],
            "rlnAngleTilt": [0, 90],
            "rlnAnglePsi": [90, 0],
            "rlnMicrographName": ["a_1", "a_2"],
            "feature": ["x", "y"],
        }
    )
    file_path = tmp_path_factory.mktemp("data") / "test.star"
    starfile.write(df, file_path)
    return file_path


@pytest.fixture(scope="session")
def mrc_file(tmp_path_factory):
    file_path = tmp_path_factory.mktemp("data") / "test.mrc"
    data = np.zeros((10, 10, 10), np.float32)
    data[5, 5, 5] = 1
    with mrcfile.new(file_path, data) as mrc:
        mrc.set_data(data)
    return file_path



from blik.reader import construct_particle_layer_tuples, get_reader


def test_reader(star_file):
    reader = get_reader(star_file)
    assert callable(reader)

    layer_data_list = reader(star_file)
    assert len(layer_data_list) == 4

    pts = layer_data_list[0]
    assert pts[2] == "points"
    vec = layer_data_list[1]
    assert vec[2] == "vectors"


def test_construct_empty_layers():
    layer_data_list = construct_particle_layer_tuples(
        coords=None,
        features=None,
        scale=1,
        exp_id="test",
    )

    assert len(layer_data_list) == 2

    pts = layer_data_list[0]
    assert pts[2] == "points"
    vec = layer_data_list[1]
    assert vec[2] == "vectors"



import numpy as np
from cryohub.utils.generic import get_columns_or_default
from cryohub.utils.types import PoseSet
from cryohub.writing.mrc import write_mrc
from cryohub.writing.star import write_star
from cryotypes.image import Image
from scipy.spatial.transform import Rotation

from .utils import invert_xyz


def write_image(path, data, attributes):
    if "experiment_id" not in attributes["metadata"]:
        raise ValueError(
            "cannot write a layer that does not have blik metadata. Add it to an experiment!"
        )
    img = Image(
        data=data,
        experiment_id=attributes["metadata"]["experiment_id"],
        pixel_spacing=attributes["scale"][0],
        stack=attributes["metadata"]["stack"],
        source=attributes["metadata"].get("source", ""),
    )
    write_mrc(img, str(path), overwrite=True)
    return [path]


def _write_particles(path, layer_data, relion_version):
    particles = []
    for data, attributes, layer_type in layer_data:
        if layer_type == "vectors":
            # vector info is actually held in particles, but this makes it
            # convenient to select everything and save
            pass
        elif "experiment_id" in attributes["metadata"]:
            data = invert_xyz(data)
            shift_cols = ["shift_z", "shift_y", "shift_x"]
            features = attributes["features"].drop(
                columns=["orientation", *shift_cols], errors="ignore"
            )

            shift = get_columns_or_default(attributes["features"], shift_cols)
            if shift is not None:
                shift = invert_xyz(shift)
                data = data - shift
            ori = get_columns_or_default(attributes["features"], "orientation")
            if ori is not None:
                ori = Rotation.concatenate(ori.reshape(-1))

            particles.append(
                PoseSet(
                    position=data,
                    shift=shift,
                    orientation=ori,
                    experiment_id=attributes["metadata"]["experiment_id"],
                    pixel_spacing=attributes["scale"][0],
                    source=attributes["metadata"].get("source", ""),
                    features=features,
                )
            )
        else:
            raise ValueError(
                "cannot write a layer that does not have blik metadata. Add it to an experiment!"
            )

    write_star(particles, path, overwrite=True, version=relion_version)
    return [path]


def write_particles_relion_30(path, layer_data):
    return _write_particles(path, layer_data, relion_version="3.0")


def write_particles_relion_31(path, layer_data):
    return _write_particles(path, layer_data, relion_version="3.1")


def write_particles_relion_40(path, layer_data):
    return _write_particles(path, layer_data, relion_version="4.0")


def write_surface_picks(path, data, attributes):
    if "experiment_id" not in attributes["metadata"]:
        raise ValueError(
            "cannot write a layer that does not have blik metadata. Add it to an experiment!"
        )

    if not str(path).endswith(".picks"):
        path = str(path) + ".picks"

    exp_id = str(attributes["metadata"]["experiment_id"])
    scale = attributes["scale"]
    surf_id = attributes["features"]["surface_id"]
    edge_color_cycle = attributes["edge_color_cycle"]
    with open(path, "wb") as f:
        np.save(f, scale)
        np.save(f, surf_id)
        np.save(f, edge_color_cycle)
        for line in data:
            np.save(f, line)
        # exp ID at the end cause it has variable length and is not a np array
        f.write(exp_id.encode())
    return [path]


def write_surface(path, data, attributes):
    if "experiment_id" not in attributes["metadata"]:
        raise ValueError(
            "cannot write a layer that does not have blik metadata. Add it to an experiment!"
        )

    if not str(path).endswith(".surf"):
        path = str(path) + ".surf"

    exp_id = str(attributes["metadata"]["experiment_id"])
    scale = attributes["scale"]
    # TODO: needs to exposed in napari
    # colormap = attributes['colormap']['colors']
    with open(path, "wb") as f:
        np.save(f, scale)
        # TODO: needs to exposed in napari
        # np.save(f, colormap)
        for d in data:
            np.save(f, d)
        # exp ID at the end cause it has variable length and is not a np array
        f.write(exp_id.encode())
    return [path]



from typing import TYPE_CHECKING

import dask.array as da
import numpy as np
from magicgui import magic_factory
from scipy.fft import fftn, fftshift

if TYPE_CHECKING:
    import napari


@magic_factory(
    auto_call=False,
    call_button="Calculate",
)
def power_spectrum(
    image: "napari.layers.Image",
) -> "napari.types.LayerDataTuple":
    """
    Power spectrum (log scale) of the image.

    First centers in real space on the origin to remove shift effects.
    """
    axes = (1, 2) if image.metadata["stack"] else None
    raw = da.compute(image.data)[0]
    power_spectrum = np.abs(
        fftshift(fftn(fftshift(raw, axes=axes), axes=axes), axes=axes)
    )
    return (
        np.log(power_spectrum + 1),
        {"name": f"{image.name} - power spectrum", "scale": image.scale},
        "image",
    )



import napari
import numpy as np
import pandas as pd
from dask.array import compute
from magicgui import magicgui
from magicgui.widgets import Container
from morphosamplers.helical_filament import HelicalFilament
from morphosamplers.sampler import (
    sample_volume_along_spline,
    sample_volume_around_surface,
)
from morphosamplers.surface_spline import GriddedSplineSurface
from scipy.spatial.transform import Rotation

from ..reader import construct_particle_layer_tuples
from ..utils import invert_xyz


def _generate_surface_grids_from_shapes_layer(
    surface_shapes, spacing_A=100, closed=False
):
    """create a new surface representation from picked surface points."""
    spacing_A /= surface_shapes.scale[0]
    colors = []
    surface_grids = []
    data_array = np.array(surface_shapes.data, dtype=object)  # helps with indexing
    for _, surf in surface_shapes.features.groupby("surface_id"):
        lines = data_array[surf.index]
        # sort so lines can be added in between at a later point
        # also move to xyz world so math is the same as reader code
        lines = [
            invert_xyz(line).astype(float)
            for line in sorted(lines, key=lambda x: x[0, 0])
        ]

        try:
            surface_grids.append(
                GriddedSplineSurface(
                    points=lines,
                    separation=spacing_A,
                    order=3,
                    closed=closed,
                )
            )
        except ValueError:
            continue

        colors.append(surface_shapes.edge_color[surf.index])

    if not colors:
        raise RuntimeError("could not generate surfaces for some reason")

    colors = np.concatenate(colors)
    return surface_grids, colors


def _resample_surfaces(image_layer, surface_grids, spacing, thickness, masked):
    volumes = []
    for surf in surface_grids:
        # transpose because zyx to xyz
        vol = sample_volume_around_surface(
            compute(image_layer.data.T)[0],
            surface=surf,
            sampling_thickness=thickness,
            sampling_spacing=spacing,
            interpolation_order=3,
            masked=masked,
        )
        volumes.append(vol.T)
    return volumes


def _generate_filaments_from_points_layer(filament_picks):
    """create a new filament representation from picked points."""
    return HelicalFilament(points=invert_xyz(filament_picks.data).astype(float))


def _resample_filament(image_layer, filament, spacing, thickness):
    return sample_volume_along_spline(
        compute(image_layer.data.T)[0],
        spline=filament,
        sampling_shape=(thickness, thickness),
        sampling_spacing=spacing,
        interpolation_order=3,
    )


@magicgui(
    labels=True,
    call_button="Generate",
    spacing_A={"widget_type": "Slider", "min": 1, "max": 500},
    inside_points={"nullable": True},
)
def surface(
    surface_shapes: napari.layers.Shapes,
    inside_points: napari.layers.Points,
    spacing_A=100,
    closed=False,
) -> napari.types.LayerDataTuple:
    """create a new surface representation from picked surface points."""
    surface_grids, colors = _generate_surface_grids_from_shapes_layer(
        surface_shapes, spacing_A
    )

    meshes = []
    exp_id = surface_shapes.metadata["experiment_id"]

    for surf in surface_grids:
        meshes.append(surf.mesh())

    offset = 0
    vert = []
    faces = []
    ids = []
    for surf_id, (v, f) in enumerate(meshes):
        f += offset
        offset += len(v)
        vert.append(v)
        faces.append(f)
        ids.append(np.full(len(v), surf_id))
    vert = np.concatenate(vert)
    faces = np.concatenate(faces)
    uniq_colors, idx = np.unique(colors, axis=0, return_index=True)
    colormap = uniq_colors[np.argsort(idx)]
    values = np.concatenate(ids) / len(colormap)
    # special case for colormap with 1 color because blacks get autoadded at index 0
    if colormap.shape[0] == 1:
        values += 1

    surface_layer_tuple = (
        (invert_xyz(vert), faces, values),
        {
            "name": f"{exp_id} - surface",
            "metadata": {
                "experiment_id": exp_id,
                "surface_grids": surface_grids,
                "surface_colors": colors,
            },
            "scale": surface_shapes.scale,
            "shading": "smooth",
            "colormap": colormap,
        },
        "surface",
    )
    return [surface_layer_tuple]


@magicgui(
    labels=True,
    call_button="Generate",
    spacing_A={"widget_type": "Slider", "min": 1, "max": 500},
)
def surface_particles(
    surface: napari.layers.Surface,
    spacing_A=50,
) -> napari.types.LayerDataTuple:
    surface_grids = surface.metadata.get("surface_grids", None)
    if surface_grids is None:
        raise ValueError("This surface layer contains no surface grid object.")
    colors = surface.metadata.get("surface_colors")

    exp_id = surface.metadata["experiment_id"]
    spacing = spacing_A / surface.scale[0]

    pos = []
    ori = []
    for surf in surface_grids:
        if not np.isclose(surf.separation, spacing):
            surf.separation = spacing
        pos.append(surf.sample())
        ori.append(surf.sample_orientations())

    pos = np.concatenate(pos)
    features = pd.DataFrame({"orientation": np.asarray(Rotation.concatenate(ori))})

    return construct_particle_layer_tuples(
        coords=invert_xyz(pos),
        features=features,
        scale=surface.scale[0],
        exp_id=exp_id,
        face_color_cycle=colors,
    )


@magicgui(
    labels=True,
    call_button="Resample",
    spacing_A={"widget_type": "Slider", "min": 1, "max": 500},
    thickness_A={"widget_type": "Slider", "min": 1, "max": 500},
)
def resample_surface(
    surface: napari.layers.Surface,
    volume: napari.layers.Image,
    spacing_A=5,
    thickness_A=200,
    masked=False,
) -> napari.types.LayerDataTuple:
    surface_grids = surface.metadata.get("surface_grids", None)
    if surface_grids is None:
        raise ValueError("This surface layer contains no surface grid object.")

    exp_id = surface.metadata["experiment_id"]
    spacing = spacing_A / surface.scale[0]
    thickness = int(np.round(thickness_A / surface.scale[0]))
    for surf in surface_grids:
        if not np.isclose(surf.separation, spacing):
            surf.separation = spacing

    vols = _resample_surfaces(volume, surface_grids, spacing, thickness, masked)

    v = napari.Viewer()
    for i, vol in enumerate(vols):
        v.add_image(
            vol,
            name=f"{exp_id} - surface_{i} resampled",
            metadata={"experiment_id": exp_id},
            scale=surface.scale,
        )


@magicgui(
    labels=True,
    call_button="Generate",
)
def filament(
    points: napari.layers.Points,
) -> napari.types.LayerDataTuple:
    filament = _generate_filaments_from_points_layer(points)

    exp_id = points.metadata["experiment_id"]

    path = filament.sample(n_samples=len(points.data) * 50)
    shapes_layer_tuple = (
        [invert_xyz(path)],
        {
            "name": f"{exp_id} - filament",
            "metadata": {
                "experiment_id": exp_id,
                "helical_filament": filament,
            },
            "scale": points.scale,
            "shape_type": "path",
        },
        "shapes",
    )
    return [shapes_layer_tuple]


@magicgui(
    labels=True,
    call_button="Generate",
    rise_A={"widget_type": "Slider", "min": 1, "max": 500},
    radius_A={"widget_type": "Slider", "min": 0, "max": 500},
    twist_deg={"widget_type": "Slider", "min": 0, "max": 360},
    twist_offset={"widget_type": "Slider", "min": 0, "max": 360},
)
def filament_particles(
    filament: napari.layers.Shapes,
    rise_A=50,
    twist_deg=0,
    twist_offset=0,
    radius_A=0,
    cyclic_symmetry_order=1,
) -> napari.types.LayerDataTuple:
    helical_filament = filament.metadata.get("helical_filament", None)
    if helical_filament is None:
        raise ValueError("This shapes layer contains no helical filament object.")

    exp_id = filament.metadata["experiment_id"]

    pos, ori = helical_filament.sample_helical(
        rise=rise_A / filament.scale[0],
        twist=twist_deg,
        radial_offset=radius_A / filament.scale[0],
        cyclic_symmetry_order=cyclic_symmetry_order,
        twist_offset=twist_offset,
        degrees=True,
    )

    features = pd.DataFrame({"orientation": np.asarray(Rotation.concatenate(ori))})

    return construct_particle_layer_tuples(
        coords=invert_xyz(pos),
        features=features,
        scale=filament.scale[0],
        exp_id=exp_id,
    )


@magicgui(
    labels=True,
    call_button="Resample",
    spacing_A={"widget_type": "Slider", "min": 1, "max": 500},
    thickness_A={"widget_type": "Slider", "min": 1, "max": 500},
)
def resample_filament(
    filament: napari.layers.Shapes,
    volume: napari.layers.Image,
    spacing_A=5,
    thickness_A=200,
) -> napari.types.LayerDataTuple:
    helical_filament = filament.metadata.get("helical_filament", None)
    if helical_filament is None:
        raise ValueError("This shapes layer contains no helical filament object.")

    exp_id = filament.metadata["experiment_id"]
    spacing = spacing_A / filament.scale[0]
    thickness = int(np.round(thickness_A / filament.scale[0]))

    vol = _resample_filament(volume, helical_filament, spacing, thickness)

    v = napari.Viewer()
    v.add_image(
        vol,
        name=f"{exp_id} - filament resampled",
        metadata={"experiment_id": exp_id},
        scale=filament.scale,
    )


class FilamentWidget(Container):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.append(filament)
        self.append(filament_particles)
        self.append(resample_filament)


class SurfaceWidget(Container):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.append(surface)
        self.append(surface_particles)
        self.append(resample_surface)



from pathlib import Path
from typing import TYPE_CHECKING, List

from magicgui import magic_factory

from ..reader import read_layers

if TYPE_CHECKING:
    import napari


@magic_factory(
    call_button="Read",
    files={"mode": "rm"},
    name_regex={"widget_type": "ListEdit"},
    names={"widget_type": "ListEdit"},
)
def file_reader(
    files: List[Path],
    name_regex: List[str],
    names: List[str],
    as_dask_array: bool = True,
) -> "napari.types.LayerDataTuple":
    """
    Read files with blik.

    name_regex: a regex string. Matching text will be used as name for the piece of data
    names: only load data matching this comma separated list of names
    """
    return read_layers(
        *files, name_regex=name_regex or None, names=names or None, lazy=as_dask_array
    )



import einops
import napari
import numpy as np
from scipy.spatial.transform import Rotation


def invert_xyz(arr):
    return arr[..., ::-1]


def generate_vectors(coords, orientations):
    """Generate basis vectors and relative colors for napari."""
    mat = Rotation.concatenate(orientations).as_matrix()
    basis_vecs = einops.rearrange(mat, "batch a b -> b batch a")
    vec_data = np.empty((len(coords) * 3, 2, 3))
    vec_color = np.empty((len(coords) * 3, 3))
    for idx, vecs in enumerate(basis_vecs):
        color = np.zeros(3)
        color[idx] = 1  # rgb
        vec_data[idx::3] = np.stack([coords, vecs], axis=1)
        vec_color[idx::3] = color
    return vec_data, vec_color


def layer_tuples_to_layers(layer_tuples):
    return [
        getattr(napari.layers, ltype.capitalize())(data, **kwargs)
        for data, kwargs, ltype in layer_tuples
    ]



# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = "0.4.1.dev20+gc4aab1a.d20230829"
__version_tuple__ = version_tuple = (0, 4, 1, "dev20", "gc4aab1a.d20230829")



from typing import TYPE_CHECKING

import numpy as np
from magicgui import magic_factory
from scipy.signal.windows import gaussian
from skimage.filters import butterworth

if TYPE_CHECKING:
    import napari


@magic_factory(
    auto_call=False,
    call_button=True,
    low={"widget_type": "FloatSlider", "min": 0, "max": 0.5},
    high={"widget_type": "FloatSlider", "min": 0, "max": 0.5},
)
def bandpass_filter(
    image: "napari.layers.Image", low: float = 0.1, high: float = 0.4
) -> "napari.types.LayerDataTuple":
    channel_axis = 0 if image.metadata["stack"] else None
    high_pass = butterworth(
        np.asarray(image.data), low, high_pass=True, channel_axis=channel_axis
    )
    low_pass = butterworth(high_pass, high, channel_axis=channel_axis)
    return low_pass, {"name": f"filtered {image.name}", "scale": image.scale}, "image"


def gaussian_kernel(size, sigma):
    window = gaussian(size, sigma)
    kernel = np.outer(window, window)
    return kernel / kernel.sum()


@magic_factory(
    auto_call=True,
    sigma={"widget_type": "FloatSlider", "min": 0.1, "max": 5, "step": 0.1},
    kernel_size={"widget_type": "Slider", "min": 3, "max": 20},
)
def gaussian_filter(
    image: "napari.layers.Image",
    sigma: float = 1,
    kernel_size: int = 3,
) -> None:
    image.interpolation2d = "custom"
    image.custom_interpolation_kernel_2d = gaussian_kernel(kernel_size, sigma)



import mrcfile
import napari
import numpy as np

from blik.widgets.file_reader import file_reader
from blik.widgets.filter import bandpass_filter, gaussian_filter
from blik.widgets.main_widget import MainBlikWidget


def test_main_widget(make_napari_viewer):
    wdg = MainBlikWidget()
    assert "test" not in wdg[0].experiment_id.choices

    viewer = make_napari_viewer()
    with napari.layers._source.layer_source(reader_plugin="blik"):
        viewer.add_image(
            np.ones((10, 10, 10)),
            name="test",
            metadata={"experiment_id": "test", "stack": False},
        )
    viewer.window.add_dock_widget(wdg)
    assert "test" in wdg[0].experiment_id.choices
    wdg[0].experiment_id.value = "test"

    # add new segmentation
    wdg[1].l_type.value = "segmentation"
    wdg[1]()
    assert viewer.layers[-1].name == "test - segmentation"

    # add new picking
    wdg[1].l_type.value = "particles"
    wdg[1]()
    assert viewer.layers[-2].name == "test - particle positions"
    assert viewer.layers[-1].name == "test - particle orientations"

    # add layer manually
    lay = viewer.add_points(name="test_points")
    wdg[2].layer.value = lay
    wdg[2]()
    assert lay.metadata["experiment_id"] == "test"


def test_reader_widget(make_napari_viewer, star_file):
    viewer = make_napari_viewer()
    wdg = file_reader()
    viewer.window.add_dock_widget(wdg)
    wdg.files.value = [star_file]
    wdg.name_regex.value = [r"\w"]

    assert len(viewer.layers) == 0
    wdg()
    assert len(viewer.layers) == 2


def test_gaussian_filter_widget(make_napari_viewer, mrc_file):
    viewer = make_napari_viewer()
    wdg = gaussian_filter()
    viewer.window.add_dock_widget(wdg)
    with mrcfile.open(mrc_file) as mrc:
        layer = viewer.add_image(
            mrc.data, metadata={"experiment_id": "test", "stack": False}
        )
    assert wdg.image.value == layer
    wdg()


def test_bandpass_filter_widget(make_napari_viewer, mrc_file):
    viewer = make_napari_viewer()
    wdg = bandpass_filter()
    viewer.window.add_dock_widget(wdg)
    with mrcfile.open(mrc_file) as mrc:
        layer = viewer.add_image(
            mrc.data, metadata={"experiment_id": "test", "stack": False}
        )
    assert wdg.image.value == layer
    wdg()
    result = viewer.layers[-1].data
    assert np.all(result != 1)



from __future__ import annotations

from typing import List

import napari
import numpy as np
import pandas as pd
from magicgui import magic_factory, magicgui
from magicgui.widgets import Container
from napari.layers import Image, Labels, Points, Shapes, Vectors
from napari.utils._magicgui import find_viewer_ancestor
from napari.utils.notifications import show_info
from scipy.spatial.transform import Rotation

from ..reader import construct_particle_layer_tuples
from ..utils import generate_vectors, layer_tuples_to_layers


def _get_choices(wdg, condition=None):
    """generate choices for the experiment_id dropdown based on the layers in the layerlist."""
    viewer = find_viewer_ancestor(wdg.native)
    if not viewer:
        return []

    choices = set()
    for lay in viewer.layers:
        if condition is not None and not condition(lay):
            continue
        try:
            exp = lay.metadata["experiment_id"]
        except KeyError:
            continue
        if exp is not None:
            choices.add(exp)
    return sorted(choices)


def _connect_points_to_vectors(p, v):
    """connect a particle points layer to a vectors layer to keep them in sync."""

    def _update_vectors():
        if not len(p.data):
            return
        if np.any(pd.isnull(p.features["orientation"])):
            obj = p.features["orientation"].astype(object)
            obj[pd.isnull(obj)] = Rotation.identity()
            p.features["orientation"] = obj
        vec_data, vec_color = generate_vectors(p.data[...], p.features["orientation"])
        v.data = vec_data
        v.edge_color = vec_color

    p.events.set_data.disconnect(_update_vectors)
    p.events.set_data.connect(_update_vectors)
    p.events.features.disconnect(_update_vectors)
    p.events.features.connect(_update_vectors)


def _connect_picking_callbacks(surf):
    @surf.bind_key("n", overwrite=True)
    def next_surface(ev):
        surf.feature_defaults["surface_id"] += 1

    @surf.bind_key("p", overwrite=True)
    def previous_surface(ev):
        surf.feature_defaults["surface_id"] -= 1


def _attach_callbacks_to_viewer(wdg):
    """attach all callbacks to the napari viewer and enable scale bar."""
    viewer = find_viewer_ancestor(wdg.native)
    if viewer:
        viewer.layers.events.inserted.connect(lambda e: _connect_layers(viewer, e))
        _connect_layers(viewer, None)

        viewer.scale_bar.unit = "0.1nm"  # pixels are 1 A
        viewer.scale_bar.visible = True


def _connect_layers(viewer, e):
    """connect all points and vectors layers with the necessary callbacks."""
    points = {}
    vectors = {}
    for lay in viewer.layers:
        if "experiment_id" not in lay.metadata:
            continue

        p_id = lay.metadata.get("p_id", None)
        if p_id is not None:
            if isinstance(lay, Points):
                points[p_id] = lay
            elif isinstance(lay, Vectors):
                vectors[p_id] = lay
        if isinstance(lay, Shapes):
            _connect_picking_callbacks(lay)

    for p_id, p in points.items():
        v = vectors.get(p_id, None)
        if v is not None:
            _connect_points_to_vectors(p, v)


@magic_factory(
    auto_call=True,
    call_button=False,
    labels=False,
    experiment_id={
        "widget_type": "ComboBox",
        "choices": _get_choices,
        "nullable": True,
    },
)
def experiment(viewer: napari.Viewer, experiment_id):
    """Select which experiment_id to display in napari and hide everything else."""
    sel = []
    if viewer is None:
        return
    for layer in viewer.layers:
        try:
            layer_exp = layer.metadata["experiment_id"]
        except KeyError:
            # leave untracked layers alone, and keep them in the selection if there
            if layer in viewer.layers.selection:
                sel.append(layer)
            continue
        if layer_exp == experiment_id:
            layer.visible = True
            sel.append(layer)
        else:
            layer.visible = False
    viewer.layers.selection = set(sel)
    experiment.current_layers = set(sel)


@magicgui(
    labels=False,
    call_button="Add",
)
def add_to_exp(layer: napari.layers.Layer):
    """add layer to the current experiment."""
    layer.metadata["experiment_id"] = add_to_exp._main_widget[
        "experiment"
    ].experiment_id.value
    if isinstance(layer, (Image, Labels)):
        if "stack" not in layer.metadata:
            layer.metadata["stack"] = False


@magicgui(
    labels=False,
    call_button="Create",
    l_type={
        "choices": ["segmentation", "particles", "surface_picking", "filament_picking"]
    },
)
def new(
    l_type,
) -> List[napari.layers.Layer]:  # noqa. This must be List, not list (magicgui...)
    """create a new layer to add to this experiment."""
    layers = getattr(new._main_widget["experiment"], "current_layers", [])
    if not layers:
        show_info("no experiment is selected")
        return []

    exp_id = new._main_widget["experiment"].experiment_id.value
    if l_type == "segmentation":
        for lay in layers:
            if isinstance(lay, Image) and lay.metadata["experiment_id"] == exp_id:
                labels = Labels(
                    np.zeros(lay.data.shape, dtype=np.int32),
                    name=f"{exp_id} - segmentation",
                    scale=lay.scale,
                    metadata={
                        "experiment_id": exp_id,
                        "stack": lay.metadata["stack"],
                    },
                )
                return [labels]
    elif l_type == "particles":
        for lay in layers:
            if lay.metadata["experiment_id"] == exp_id:
                layers = construct_particle_layer_tuples(
                    coords=None, features=None, scale=lay.scale[0], exp_id=exp_id
                )
                return layer_tuples_to_layers(layers)
    elif l_type == "surface_picking":
        for lay in layers:
            if isinstance(lay, Image) and lay.metadata["experiment_id"] == exp_id:
                pts = Shapes(
                    name=f"{exp_id} - surface lines",
                    edge_width=50 / lay.scale[0],
                    scale=lay.scale,
                    metadata={"experiment_id": exp_id},
                    features={"surface_id": np.empty(0, int)},
                    feature_defaults={"surface_id": 0},
                    edge_color_cycle=np.random.rand(30, 3),
                    edge_color="surface_id",
                    ndim=3,
                )

                return [pts]
    elif l_type == "filament_picking":
        for lay in layers:
            if isinstance(lay, Image) and lay.metadata["experiment_id"] == exp_id:
                pts = Points(
                    name=f"{exp_id} - filament picks",
                    size=20 / lay.scale[0],
                    scale=lay.scale,
                    metadata={"experiment_id": exp_id},
                    face_color_cycle=np.random.rand(30, 3),
                    ndim=3,
                )

                return [pts]

    show_info(f"cannot create a new {l_type}")
    return []


@magicgui(
    labels=False,
    auto_call=True,
    thickness_A={"widget_type": "FloatSlider", "min": 0, "max": 500},
)
def slice_thickness_A(viewer: napari.Viewer, thickness_A=0):
    viewer.dims.thickness = (thickness_A,) * viewer.dims.ndim


class MainBlikWidget(Container):
    """
    Main widget for blik controls.

    Allows to select which layers to view based on the experiment id, to add
    existing layer to a certain experiment id, and to create new analysis layers
    within.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        exp = experiment()
        self.parent_changed.connect(lambda _: _attach_callbacks_to_viewer(exp))
        self.append(exp)
        self.append(new)
        self.append(add_to_exp)
        # self.append(slice_thickness_A)  # only available in napari 0.5.0

    def append(self, item):
        super().append(item)
        item._main_widget = self
